ERROR_RUNTIME_ERROR = "wamp.error.runtime_error"
ERROR_INVALID_ARGUMENT = "wamp.error.invalid_argument"
ERROR_INTERNAL_ERROR = "wamp.error.internal_error"
CLOSE_REALM = "wamp.close.close_realm"
CLOSE_GOODBYE_AND_OUT = "wamp.close.goodbye_and_out"
